module Labyrinth
module QCutList
module CutList
class PartList
  def initialize()
    @parts = Array.new
    @partCount = Array.new
  end  ### Constructor
  def parts
    @parts
  end ###def parts
  def partCount
    @partCount
  end ###def partCount
  def add(inPart)
    index = @parts.index(inPart)
    if(index != nil)
      @partCount[index] = @partCount[index] + 1
    else
      @parts.push(inPart)
      @partCount.push(1)
    end  ###if
  end ###def add
  def empty?
    @parts.length == 0
  end
  def getList
    @parts
  end
end ###Class PartList
class CutListPart
  attr_accessor :copyIndex, :copyCount
  def initialize(c, name, subAssemblyName, materialName, mainFace, metricVolume, respectGrain, ancestors)
    boundingBox = c.bounds
    if c.respond_to? "definition"
     boundingBox = c.definition.bounds
    else
      if c.is_a?(Sketchup::Group)
        group_definition = CutList::group_definition(c)
        boundingBox = group_definition.bounds
      end
    end
    width = boundingBox.width
    height = boundingBox.height
    depth = boundingBox.depth
    ancestors.push(c)
    indexAncestor = ancestors.count - 1
    while indexAncestor >= 0
      ancestor = ancestors[indexAncestor]
      if ancestor.is_a?( Sketchup::ComponentInstance) || ancestor.is_a?( Sketchup::Group)
        trans=ancestor.transformation.to_a
        scalex=Math.sqrt(trans[0]**2+trans[1]**2+trans[2]**2)
        scaley=Math.sqrt(trans[4]**2+trans[5]**2+trans[6]**2)
        scalez=Math.sqrt(trans[8]**2+trans[9]**2+trans[10]**2)
        width = width * scalex
        height = height * scaley
        depth = depth * scalez
      end
      indexAncestor -= 1
    end
    ancestors.pop
    sizes = getSortedArray([width,height,depth])
    useGrain = false
    if respectGrain
      materialInfo = Materials.getMaterial(materialName)
      sheetGrainDir = materialInfo['grainDir']
      if sheetGrainDir == 'v' || sheetGrainDir == 'h'
        if mainFace.material
          if !Materials.isMaterialRotatable?(mainFace)
            go = 0
          else
            go = Materials.isMaterialRotated(c) ? 2 : 1
          end
        else
          go = 1
        end
        if go != 0
          useGrain = true
        end
      end
    end
    @thickness = sizes[0].inch
    if useGrain
      realTrans = c.transformation
      c.transformation = Geom::Transformation.new([0,0,0])
      grainAxis = c.transformation.xaxis
      testAxis = X_AXIS
      if @thickness != depth
        if @thickness == width
          width = depth
          grainAxis = c.transformation.yaxis
        else
          height = depth
        end
      end
      angleInDegs = grainAxis.angle_between(X_AXIS).radians # - rotx
      isComponentRotated = (angleInDegs >= 75 && angleInDegs < 105) || (angleInDegs >= 255 && angleInDegs < 285)
      if isComponentRotated #&& false
        if go == 1
          go = 2
        else
          go = 1
        end
      end
      if go != 2
        @length = width
        @width = height
      else
        @length = height
        @width = width
      end
      c.transformation = realTrans
    else
      @length = sizes[2].inch
      @width = sizes[1].inch
    end
    @lengthInFeet = @length.to_feet
    dimCalculations()
    @mainFace = mainFace
    @materialName = materialName
    @name = strip(name,@length.to_s, @width.to_s, @thickness.to_s )
    @subAssemblyName = strip(subAssemblyName, @length.to_s, @width.to_s, @thickness.to_s )
    @canRotate = true
    @metricVolume = metricVolume
    @metric = CutList.metricModel?
    @locationOnBoard = nil
    @copyIndex = 0
    @copyCount = 1
  end
  def swapLengthWidth
    length = @length
    @length = @width
    @width = length
    if @locationOnBoard
      x = @locationOnBoard[0]
      @locationOnBoard[0] = @locationOnBoard[1]
      @locationOnBoard[1] = x
    end
  end
  def dimCalculations
    @area = @length*@width
    @volume = @area*@thickness
    @squareFeet = @area/144
    @boardFeet = @volume/144
    @length_px = CutList::float_round_to(0,((@length/12)*100).to_f)
    @width_px = CutList::float_round_to(0,((@width/12)*100).to_f)
  end
  def getSortedArray(array)
    size = array.size()
    pass = size
    for i in (0..pass-2)
      for j in (0..pass-2)
        if (array[j+1] < array[j])
          tmp = array[j]
          array[j] = array[j+1]
          array[j+1] = tmp
        end
      end
    end
    return array
  end
  def strip(name,inV1,inV2,inV3)
    val=name
    if(name=="noname")
      val=("noname_"+inV1+inV2+inV3)
      val = val.gsub(/[ ]/, "_")
      val = val.gsub(/['~"]/, "")
      val = val.gsub(/[\/]/, "-")
    end ## end if
    return val
  end  ## end strip
  def getTotalLength
    if @metricVolume
      CutList::float_round_to(4,(@lengthInFeet*0.3048))
    else
      CutList::float_round_to(2,@lengthInFeet)
    end
  end
  def getLengthString
      @length.to_l.to_s
  end
  def getWidthString
      @width.to_l.to_s
  end
  def getWidth
      @width
  end
  def getLength
    @length
  end
  def getThicknessString
      @thickness.to_l.to_s
  end
  def getThickness
    @thickness
  end
  def getMaterial
    @materialName
  end
  def getName
    @name
  end
  def getSubAssemblyName
    @subAssemblyName
  end
  def getBoardFeet
    if @metricVolume
      CutList::float_round_to(6,(@boardFeet*(2359.7372232207958956904236222001/1000000)))
    else
      CutList::float_round_to(2,@boardFeet)
    end
  end
  def getBoardFeetLabel
    if @metricVolume
      return "Cubic m"
    else
      return "Board Feet"
    end
  end
  def getSquareFeet
    if @metricVolume
      CutList::float_round_to(6,(@squareFeet*(0.092903040189522201889968968842358)))
    else
      CutList::float_round_to(2,@squareFeet)
    end
  end
  def getCanRotate
    @canRotate
  end
  def getLengthPx
    @length_px
  end
  def getWidthPx
    @width_px
  end
  def addLocationOnBoard(coords)
    @locationOnBoard = coords
  end
  def getLocationOnBoard
    @locationOnBoard
  end
  def getLocationOnBoardInPx
    Array[CutList::float_round_to(0,((@locationOnBoard[0]/12)*100).to_f), CutList::float_round_to(0,((@locationOnBoard[1]/12)*100).to_f)]
  end
  def changeWidth(width)
    @width = width.inch
    dimCalculations()
  end
  def changeThickness(thickness)
    @thickness = thickness.inch
    dimCalculations()
  end
  def convertMeasureForCLP(measure)
    if CutList::modelInMeters?
      measure = CutList::float_round_to(4,measure.to_f)
    end
    return measure      
  end
  def deep_clone
   Marshal::load(Marshal.dump(self))  
  end 
  def summary
    return (getName + " (" + getLengthString + ", " + getWidthString + ", " + getThicknessString + ") " + getMaterial)
  end
end
class SheetPart < CutListPart
  def getBoardFeet
    getSquareFeet
  end
  def getBoardFeetLabel
    if @metricVolume
      return "Square m"
    else
      return "Square feet"
    end
  end
end
class CutListPartList
  def initialize()
    @componentList = Array.new
    @componentListByMaterialAndThickness = Hash.new
  end 
  def addToPartDatabase(cutListPart)
    return if cutListPart == nil
    materialName = cutListPart.getMaterial
    thickness = cutListPart.getThickness.inch.to_s
    if !@componentListByMaterialAndThickness.include? materialName
      newThicknessHash = Hash.new
      @componentListByMaterialAndThickness[materialName] = newThicknessHash
    end
    if !@componentListByMaterialAndThickness[materialName].include? thickness
      newPartArray = Array.new
      @componentListByMaterialAndThickness[materialName][thickness] = newPartArray
    end
    @componentListByMaterialAndThickness[materialName][thickness].push( cutListPart )
  end
  def removeFromPartDatabase(cutListPart)
    return if cutListPart == nil
    materialName = cutListPart.getMaterial
    thickness = cutListPart.getThickness.inch.to_s
    @componentListByMaterialAndThickness[materialName][thickness].delete( cutListPart )
    if @componentListByMaterialAndThickness[materialName][thickness].empty?
      @componentListByMaterialAndThickness[materialName].delete(thickness)
    end
    if @componentListByMaterialAndThickness[materialName].empty?
      @componentListByMaterialAndThickness.delete(materialName)
    end
  end
  def add(cutListPart)
    @componentList.push(cutListPart)
    addToPartDatabase(cutListPart)
  end
  def remove(cutListPart)
    removeFromPartDatabase(cutListPart)
    @componentList.delete(cutListPart)
  end
  def empty?
    @componentList.length == 0
  end
  def count
    @componentList.length
  end
  def allPartsSameMaterial?
    return true if @componentList.empty?
    materialName = @componentList.first.getMaterial
    return ( @componentList.select{|x| x.getMaterial != materialName } == nil )
  end
  def allPartsSameThickness?
    return true if @componentList.empty?
    thickness = @componentList.first.getThickness
    return (@componentList.select{|x| x.getThickness != thickness } == nil )
  end
  def getList
      return @componentList
  end
  def removeFirst!
    cutListPart = @componentList.shift
    removeFromPartDatabase(cutListPart)
    return cutListPart
  end
  def insertFirst!(cutListPart)
    @componentList.unshift(cutListPart)
    addToPartDatabase(cutListPart)
  end
  def sort
    sortedComponentList = @componentList
    size = sortedComponentList.size()
    pass = size
    for i in (0..pass-2)
      for j in (0..pass-2)
        if (sortedComponentList[j].getBoardFeet < sortedComponentList[j+1].getBoardFeet)
          tmp = sortedComponentList[j+1]
          sortedComponentList[j+1] = sortedComponentList[j]
          sortedComponentList[j] = tmp
        elsif (sortedComponentList[j].getBoardFeet == sortedComponentList[j+1].getBoardFeet)
          if (sortedComponentList[j].getTotalLength < sortedComponentList[j+1].getTotalLength )
            tmp = sortedComponentList[j+1]
            sortedComponentList[j+1] = sortedComponentList[j]
            sortedComponentList[j] = tmp
          elsif (sortedComponentList[j].getTotalLength == sortedComponentList[j+1].getTotalLength )
            if (sortedComponentList[j].getName > sortedComponentList[j+1].getName )
              tmp = sortedComponentList[j+1]
              sortedComponentList[j+1] = sortedComponentList[j]
              sortedComponentList[j] = tmp
            end
          end
        end
      end
    end
    return sortedComponentList
  end
  def makeCopies(count, orderBy='byAssembly')
    return nil if count < 2
    @componentListByMaterialAndThickness.each { |matName, matValue|
      matValue.each { |thicknessName, thicknessValue|
        if orderBy == 'byPart'
          parts = []
          thicknessValue.each { |cutListPart|
            cutListPart.copyCount = count
            parts.push(cutListPart)
            (1...count).each { |index|
              partNew = cutListPart.clone;
              partNew.copyIndex = index
              parts.push(partNew)
            }
          }
          @componentListByMaterialAndThickness[matName][thicknessName] = parts
        else # "byAssembly"
          parts = []
          list = @componentListByMaterialAndThickness[matName][thicknessName];
          (1...count).each { |index|
            thicknessValue.each { |cutListPart|
              cutListPart.copyCount = count
              partNew = cutListPart.clone;
              partNew.copyIndex = index
              parts.push(partNew)
            }
          }
          list.concat(parts)
        end
      }
    }
  end
  def sortByName
    sortedComponentList = @componentList
    size = sortedComponentList.size()
    pass = size
    for i in (0..pass-2)
	for j in (0..pass-2)
		if (sortedComponentList[j].getName > sortedComponentList[j+1].getName )
			tmp = sortedComponentList[j+1]
			sortedComponentList[j+1] = sortedComponentList[j]
			sortedComponentList[j] = tmp
		end
	end
    end
  end
  def sort!
    sortedList = @componentList.sort
    @componentList = sortedList
  end
  def deep_clone
    dup
  end
  def splitPartsListByMaterial
    @listOfPartsListsByMaterial = Array.new
    if allPartsSameMaterial?
        @listOfPartsListsByMaterial.push(self)
        return @listOfPartsListsByMaterial 
    end
    @componentListByMaterialAndThickness.each{ |material,thickness|
      partListByMaterial = CutListPartList.new
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          partListByMaterial.add(part)
        }
      }
      @listOfPartsListsByMaterial.push(partListByMaterial)
    }
    return @listOfPartsListsByMaterial
  end
  def splitPartsListByThicknessAndMaterial
    @listOfPartsListsByMaterialAndThickness = Array.new
    if allPartsSameMaterial? && allPartsSameThickness?
        @listOfPartsListsByMaterialAndThickness.push(self)
        return @listOfPartsListsByMaterialAndThickness
    end
    @componentListByMaterialAndThickness.each{ |material,thickness|
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        partListByThicknessAndMaterial = CutListPartList.new
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          partListByThicknessAndMaterial.add(part)
        }
        @listOfPartsListsByMaterialAndThickness.push(partListByThicknessAndMaterial)
      }
    }
    return @listOfPartsListsByMaterialAndThickness
  end
  def splitPartsListByThickness
    @listOfPartsListsByThickness = Array.new
    if allPartsSameThickness?
        @listOfPartsListsByThickness.push(self)
        return @listOfPartsListsByThickness
    end
    thicknessHash = Hash.new
    @componentListByMaterialAndThickness.each{ |material,thickness|
      @componentListByMaterialAndThickness[material].each{ |thickness,part|
        if !thicknessHash.include? thickness
          partListByThickness = CutListPartList.new
          thicknessHash[thickness] = partListByThickness
        end
        @componentListByMaterialAndThickness[material][thickness].each{ |part|
          thicknessHash[thickness].add(part)
        }
      }
    }
    thicknessHash.each { |thickness,partList| @listOfPartsListsByThickness.push(partList) }
    return @listOfPartsListsByThickness
  end 
  def +(cutListPartList)
    return if cutListPartList == nil
    cutListPartList.getList.each { |part| add(part) }
  end
end
class SolidPartList < CutListPartList
end
class SheetPartList < CutListPartList
end
class BoardList < CutListPartList
end
end # CutList
end # QCutList
end # Labyrinth
