module Labyrinth
module QCutList
module CutList
  class Reporter
    cutlist_Default_Options = {
      :compactList => true,
      :listAllSorted => false,
      :linearFeet => true,
      :outFileUsed => true,
      :outFileName => "CutList.csv",
      :clpFileUsed => true,
      :clpFileName => "CutListPlusImport.csv",
      :printPage => false,
      :showComps => true,
      :showSheet => true,
      :showParts => true,
      :forceBoardFeet => false,
      :layout => false,
      :su5 => false,
      :partWords => ["part"],
      :sheetWords => ["sheet"],
      :svgLayout => false
    }
    layout_Default_Options ={
      :fourq => false,
      :fiveq => false,
      :sixq => false,
      :eightq => false,
      :tenq => false,
      :nominalMargin => 0,
      :nominalOut => false,
      :board4w => false,
      :board6w => true,
      :board8w => false,
      :board10w => false,
      :board12w => false,
      :board2l => false,
      :board4l => false,
      :board6l => false,
      :board8l => true,
      :board10l => false,
      :board12l => false,
      :sheet2w => false,
      :sheet4w => true,
      :sheet5w => false,
      :sheet2l => false,
      :sheet4l => false,
      :sheet5l => false,
      :sheet8l => true,
      :nominalWidth => true,
      :kerfSize => 2.inch/16,
      :splitWideParts => true,
      :splitThickParts => false,
      :layoutRuleA => true,
      :layoutRuleB => true,
      :layoutByLength => false,
      :layoutByMaterial => false,
      :displayUnplacedParts => true,
      :useSawKerf => true,
      :sawKerfSize => 0,
      :sawKerfUnits => "8th",
      :copies => 1,
      :copyOrderBy => "byAssembly"
    }
    @@options = {
      :cutlist_Options => cutlist_Default_Options,
      :layout_Options => layout_Default_Options
    }
    def sketchupInit(cutlist_options, layout_options)
      @@options[:layout_Options].merge!(layout_options)
      @@options[:cutlist_Options].merge!(cutlist_options)
      @metric = CutList::metricModel?
      @askFirstTime = false
      @volumeMeasureInMetric = (@metric && !@@options[:cutlist_Options][:forceBoardFeet])
      setBoardOptions()
      setSheetOptions()
      @inBoardList = BoardList.new()
      main()
    end
    def setBoardOptions
      setBoardThicknessOptions
      setBoardWidthOptions
      setBoardLengthOptions
      @boardOptions = [@boardLengthOptions, @boardWidthOptions, @quarter, @@options[:layout_Options][:nominalWidth]]
    end
    def setSheetOptions
      setSheetWidthOptions
      setSheetLengthOptions
      @sheetQuarter = [false, false, false, false, false]
      @sheetOptions = [@sheetLengthOptions, @sheetWidthOptions, @sheetQuarter, true]
    end
    def setBoardThicknessOptions
      @quarter = [@@options[:layout_Options][:fourq],
                  @@options[:layout_Options][:fiveq],
                  @@options[:layout_Options][:sixq],
                  @@options[:layout_Options][:eightq],
                  @@options[:layout_Options][:tenq]]
    end
    def setBoardWidthOptions
      @boardWidthOptions = Array.new
      @boardWidthOptions.push(4) if @@options[:layout_Options][:board4w]
      @boardWidthOptions.push(6) if @@options[:layout_Options][:board6w]
      @boardWidthOptions.push(8) if @@options[:layout_Options][:board8w]
      @boardWidthOptions.push(10) if @@options[:layout_Options][:board10w]
      @boardWidthOptions.push(12) if @@options[:layout_Options][:board12w]
    end
    def setBoardLengthOptions
      @boardLengthOptions = Array.new
      @boardLengthOptions.push(2*12) if @@options[:layout_Options][:board2l]
      @boardLengthOptions.push(4*12) if @@options[:layout_Options][:board4l]
      @boardLengthOptions.push(6*12) if @@options[:layout_Options][:board6l]
      @boardLengthOptions.push(8*12) if @@options[:layout_Options][:board8l]
      @boardLengthOptions.push(10*12) if @@options[:layout_Options][:board10l]
      @boardLengthOptions.push(12*12) if @@options[:layout_Options][:board12l]
    end
    def setSheetWidthOptions
      @sheetWidthOptions = Array.new
      if (CutList::metricModel?)
        @sheetWidthOptions.push(610/25.4) if @@options[:layout_Options][:sheet2w]
        @sheetWidthOptions.push(1220/25.4) if @@options[:layout_Options][:sheet4w]
        @sheetWidthOptions.push(1525/25.4) if @@options[:layout_Options][:sheet5w]
      else
        @sheetWidthOptions.push(2*12) if @@options[:layout_Options][:sheet2w]
        @sheetWidthOptions.push(4*12) if @@options[:layout_Options][:sheet4w]
        @sheetWidthOptions.push(5*12) if @@options[:layout_Options][:sheet5w]
      end
    end
    def setSheetLengthOptions
      @sheetLengthOptions = Array.new
      if (CutList::metricModel?)
        @sheetLengthOptions.push(610/25.4) if @@options[:layout_Options][:sheet2l]
        @sheetLengthOptions.push(1220/25.4) if @@options[:layout_Options][:sheet4l]
        @sheetLengthOptions.push(1525/25.4) if @@options[:layout_Options][:sheet5l]
        @sheetLengthOptions.push(2440/25.4) if @@options[:layout_Options][:sheet8l]
      else
        @sheetLengthOptions.push(2*12) if @@options[:layout_Options][:sheet2l]
        @sheetLengthOptions.push(4*12) if @@options[:layout_Options][:sheet4l]
        @sheetLengthOptions.push(5*12) if @@options[:layout_Options][:sheet5l]
        @sheetLengthOptions.push(8*12) if @@options[:layout_Options][:sheet8l]
      end
    end
    def main
      if (components() != nil)
        if (@@options[:cutlist_Options][:layout] || @@options[:cutlist_Options][:svgLayout])
          layout()
        end
        output()
      end
    end
    def layout
      Sketchup::set_status_text(("CutList layout generation..."), SB_PROMPT)
      Sketchup::set_status_text(" ", SB_VCB_LABEL)
      Sketchup::set_status_text(" ", SB_VCB_VALUE)
      @unplacedPartsList = CutListPartList.new
      if !@sheetPartList.empty? && @@options[:cutlist_Options][:showSheet]
        sheetPartList = @sheetPartList.deep_clone
        sheetPartPreParser = SheetPartPreParser.new(sheetPartList, @inBoardList, @sheetOptions, @@options[:layout_Options], @volumeMeasureInMetric)
        @listOfSheetPartsLists, unplacedPartsList = sheetPartPreParser.run
        @unplacedPartsList +(unplacedPartsList) if unplacedPartsList != nil
        @layoutSheets = Array.new
        @listOfSheetPartsLists.each { |sheetPartList|
          sheetPartLayoutEngine = BestFitLayoutEngine.new(sheetPartList, @inBoardList, @sheetOptions, @@options[:layout_Options], @volumeMeasureInMetric, @availableSheets, @edgeWaste)
          layoutSheets, unplacedPartsList = sheetPartLayoutEngine.run
          @layoutSheets += layoutSheets if layoutSheets != nil
          @unplacedPartsList +(unplacedPartsList) if unplacedPartsList != nil
        }
      end
      Sketchup::set_status_text((""), SB_PROMPT)
    end
    def getSubComponents(entityList, level, subAssemblyName)
      model = Sketchup.active_model
      selection = model.selection
      levelHasComponents = false
      for c in entityList
        inSelection = selection.contains? c
        if ((inSelection || level>1) && c.layer.visible? && !c.hidden?)
          if ((c.is_a? (Sketchup::ComponentInstance)) || (c.is_a? (Sketchup::Group)))
            compName = nil
            if c.is_a? (Sketchup::ComponentInstance)
              compName = c.definition.name
            elsif c.is_a? (Sketchup::Group)
              compName = c.name
              if (compName == nil || compName == "")
                compName = CutList.group_definition(c).name
              end
            end ##if
            partMaterial = nil
            mainFace = Materials.getMainFace(c)
            if mainFace
              partMaterial = mainFace.material.name if mainFace.material
              if partMaterial == nil
                partMaterial = c.material.name if c.material
              end
            end
            if (partMaterial && isPartOrSheet(@@options[:cutlist_Options][:partWords], partMaterial)) ||
              isPartOrSheet(@@options[:cutlist_Options][:partWords], compName)
              @partList.add(compName)
              levelHasComponents = true
              next #move on to the next part at this level
            end
            subList = nil
            if c.is_a? (Sketchup::ComponentInstance)
              subList = c.definition.entities
            elsif c.is_a? (Sketchup::Group)
              subList = c.entities
            end ##if
            @ancestors.push(c)
            hasSubComponents = getSubComponents(subList, level+1, compName)
            if (!hasSubComponents)
              name = " "+compName
              if (name== " ")
                name = "noname"
              end
              material = nil
              mainFace = Materials.getMainFace(c)
              if mainFace
                material = mainFace.material.name if mainFace.material
                if material == nil
                  material = c.material.name if c.material
                  if material == nil
                    indexAncestor = @ancestors.count - 1
                    while material == nil && indexAncestor >= 0
                      ancestor = @ancestors[indexAncestor]
                      if ancestor.is_a?( Sketchup::ComponentInstance) || ancestor.is_a?( Sketchup::Group)
                        material = ancestor.material.name if ancestor.material
                      end
                      indexAncestor -= 1
                    end
                  end
                end
              end
              if (material && isPartOrSheet(@@options[:cutlist_Options][:sheetWords], material))
                sheetPart = SheetPart.new(c, name, subAssemblyName, material, mainFace, @volumeMeasureInMetric, @@options[:layout_Options][:respectGrain], @ancestors)
                @sheetPartList.add(sheetPart)
              end ##if
            else
            end
            levelHasComponents = true
          end
        end #if
      end #for c
      @ancestors.pop
      return levelHasComponents
    end
    def isPartOrSheet(inList, compName)
      if !compName
        return false
      end
      compName_lc = compName.downcase
      @found = false
      @exclusionFound = false
      inList.each do |p|
        matchWord = p
        exclude = ((p =~ /^-/) != nil)
        next if ($' == '' && exclude)
        matchWord = $' if exclude
        if compName_lc.index(matchWord) != nil
          if (exclude)
            @exclusionFound = true
            @found = false
          end
          if (!@exclusionFound)
            @found = true
          end
        end
      end # end do loop
      return @found
    end
    def invert_selection!(model)
      ss = model.selection
      model.entities.each { |e| ss.toggle(e) }
      return model
    end
    def components
      @partList = PartList.new()
      @solidPartList = SolidPartList.new
      @sheetPartList = SheetPartList.new
      model = Sketchup.active_model
      selection = model.selection
      if (selection.empty?)
        return nil
      end
      entities = model.entities
      @mname = model.title
      Sketchup::set_status_text(("CutList component discovery..."), SB_PROMPT)
      Sketchup::set_status_text(" ", SB_VCB_LABEL)
      Sketchup::set_status_text(" ", SB_VCB_VALUE)
      @ancestors = []
      getSubComponents(selection, 1, @mname)
      if (@solidPartList.empty? && @sheetPartList.empty? && @partList.empty?)
        UI.beep
        UI.messagebox("No Components were found in your model where the material of the component matched a selected sheet goods material.\nYou must create a Component from your selection.\nClick on Help for more info.\nNo Cutlist will be generated.")
        Sketchup::set_status_text((""), SB_PROMPT)
        return nil
      end
      @sheetPartList.makeCopies(@@options[:layout_Options][:copies], @@options[:layout_Options][:copyOrderBy])
      if @@options[:cutlist_Options][:listAllSorted]
        @solidPartList.sort
        @sheetPartList.sort
      end
      Sketchup::set_status_text((""), SB_PROMPT)
      return true
    end
    def output
    end #def output
  end #class Reporter
end # CutList
end # QCutList
end # Labyrinth
